import type { GameData } from "$lib/GameData";
import { getDiceRoll } from "$lib/server/getDiceRoll";
import type { State } from "$lib/State";
import { hasOnlyKeys, hasProperty } from "$lib/validation";

export const FIRST_ROLL_LOST = -1;
export const FIRST_ROLL_PENDING = 0;
export const GAME_SCORE_THRESHOLD = 10_000;

export enum GameEventKind {
	SeatPlayers = "SeatPlayers",
	RollForFirst = "RollForFirst",
	Roll = "Roll",
	Hold = "Hold",
	Score = "Score",
}

export interface GameEventData {
	kind: string;
	player?: number;
	value?: number | number[];
}

export interface GameEvent extends GameEventData {
	run: (state: State) => void;
}

export function isGameEventData(target: unknown): target is GameEvent {
	if (typeof target !== "object") {
		return false;
	}

	if (!hasProperty(target, "kind", "string")) {
		return false;
	}

	// TODO: add checks to make sure that, if it has optional properties, they
	//       are the correct type.
	return hasOnlyKeys(target, ["kind", "player", "value"]);
}

export function getGameEvent(data: GameData, event: GameEventData): GameEvent {
	switch (event.kind) {
		case GameEventKind.SeatPlayers:
			if (event.value !== data.players.length) {
				throw new Error("must seat all of the players in the game");
			}

			return new SeatPlayers(event);
		case GameEventKind.RollForFirst:
			return new RollForFirst(event);
		case GameEventKind.Roll:
			// Obviously a client can't send the roll they want to make. Instead, the
			// client sends a Roll event with a value representing the number of dice
			// they intend to roll, and then the server creates an array of that length
			// with random dice values.
			if (typeof event.value !== "number") {
				throw new Error("event must include the number of dice to roll");
			}

			event.value = getDiceRoll(event.value, Math.random);

			return new Roll(event);
		case GameEventKind.Hold:
			return new Hold(event);
		case GameEventKind.Score:
			return new Score(event);
		default:
			throw new Error(`${event.kind} is not a valid kind of event`);
	}
}

/**
 * SeatPlayers takes a value which represents the number of players who will be playing.
 * It initializes the scores for that number of players.
 */
export class SeatPlayers implements GameEvent {
	kind: GameEventKind.SeatPlayers;
	value: number;

	constructor(event: GameEventData) {
		const value = throwIfNoValueNumber(event);

		this.kind = GameEventKind.SeatPlayers;
		this.value = value;
	}

	run(state: State) {
		throwIfGameOver(state);

		if (state.scores !== undefined) {
			throw new Error("players already seated!");
		}

		state.scores = new Array(this.value).fill(FIRST_ROLL_PENDING);
	}
}

/**
 * RollForFirst takes a player index and a value which represents the pips on the die
 * that the player rolled. It represents an attempt to roll the highest die and go
 * first.
 */
export class RollForFirst implements GameEvent {
	kind: GameEventKind.RollForFirst;
	player: number;
	value: number;

	constructor(event: GameEventData) {
		this.kind = GameEventKind.RollForFirst;

		const player = throwIfNoPlayer(event);
		const value = throwIfNoValueNumber(event);

		this.player = player;
		this.value = value;
	}

	run(state: State) {
		const scores = state.scores ?? [];

		throwIfGameOver(state);

		if (state.playing !== undefined) {
			throw new Error("first player has already been selected");
		}

		if (scores.length <= this.player) {
			throw new Error("this player index is out of bounds");
		}

		if (scores[this.player] !== FIRST_ROLL_PENDING) {
			throw new Error("this player has already rolled");
		}

		scores[this.player] = this.value;

		let best = 0;
		let winningIndex = -1;
		const ties = new Set<number>();

		// check for a winner
		for (let i = 0; i < scores.length; i++) {
			const score = scores[i];

			// if someone hasn't rolled, no winner yet
			if (score === 0) {
				best = 0;
				winningIndex = -1;
				ties.clear();
				break;
			}

			if (score > best) {
				// This is the new high score...
				best = score;
				winningIndex = i;
				ties.clear();
			} else if (score === best) {
				// ...or it is tied for the best score.
				ties.add(winningIndex);
				ties.add(i);
			}
		}

		if (winningIndex > -1) {
			// If a winner was found...
			if (ties.size < 1) {
				// ...and there are no ties...
				state.playing = winningIndex;
				state.scores = scores.fill(0);
				state.dieCount = 6;
			} else {
				// ...otherwise, setup for tie breaking rolls.
				state.scores = scores.map((_, i) =>
					ties.has(i) ? FIRST_ROLL_PENDING : FIRST_ROLL_LOST,
				);
			}
		}
	}
}

/**
 * Roll takes a player index and a value which is an array of numbers representing the
 * number of pips on each rolled die. The Roll event represents a player rolling however
 * many dice they have available.
 */
export class Roll implements GameEvent {
	kind: GameEventKind.Roll;
	player: number;
	value: number[];

	constructor(event: GameEventData) {
		this.kind = GameEventKind.Roll;

		const player = throwIfNoPlayer(event);
		const value = throwIfNoValueArray(event);

		this.player = player;
		this.value = value;
	}

	run(state: State) {
		throwIfGameOver(state);
		throwIfWrongTurn(state, this.player);

		if (state.dieCount !== this.value.length) {
			throw new Error("player is rolling the wrong number if dice");
		}

		if (state.dice !== undefined) {
			throw new Error("player has already rolled");
		}

		state.dice = this.value;
		delete state.dieCount;
	}
}

/**
 * Hold takes a player index and a value which is an array of numbers representing
 * indexes of rolled dice that a player wants to hold. Hold represents the part of a
 * player's turn where they select scoring dice to hold onto.
 *
 * If the player holds all the dice, the number of dice they can roll will reset to 6,
 * and they will be required to roll again (a player must always roll when they have 6)
 * dice available.
 */
export class Hold implements GameEvent {
	kind: GameEventKind.Hold;
	player: number;
	value: number[];

	constructor(event: GameEventData) {
		this.kind = GameEventKind.Hold;

		const player = throwIfNoPlayer(event);
		const value = throwIfNoValueArray(event);

		this.player = player;
		this.value = value;
	}

	run(state: State) {
		const { dice } = state;

		throwIfGameOver(state);
		throwIfWrongTurn(state, this.player);

		if (!dice) {
			throw new Error("player hasn't rolled yet");
		}

		if (this.value.length < 1) {
			throw new Error("player cannot hold 0 dice");
		}

		// heldValues represents the actual values of the held dice based on the indexes
		// of the held dice.
		const heldValues = this.value.map((val) => {
			if (dice[val] === undefined) {
				throw new Error("player is trying to hold non-existent dice");
			}

			return dice[val];
		});

		const counts = new Map<number, number>();
		let total = 0;

		const valueSet = new Set(heldValues);
		if (valueSet.size === 6) {
			// Detect a run of six: if the held values create a set of values with a size
			// of six, then, since there are only six dice, it MUST be a run of six.
			total = 2_000;
		} else if (valueSet.size === 2 && this.value.length === 6) {
			// Detect two threes of a kind: if the number of held values was six, and the
			// set of unique values is two, then there MUST have been two threes of a
			// kind.
			total = 1_600;
		} else {
			// A player can use a "push" if they are using every one of their rolled
			// dice.
			let canPush = this.value.length === dice.length;

			for (const val of heldValues) {
				const count = counts.get(val) ?? 0;
				counts.set(val, count + 1);
			}

			for (let pips = 1; pips <= 6; pips++) {
				if (!counts.has(pips)) continue;

				// This for sure exists because of the check at t he top of the loop.
				const count = counts.get(pips)!;
				const pipsScore = scorePips(count, pips);

				if (!pipsScore) {
					// The player can only hold scoring dice...
					if (!canPush || count !== 2) {
						throw new Error("player is holding non-scoring dice");
					} else {
						// ...unless they have a push.
						canPush = false;
					}
				}

				total += pipsScore;
			}
		}

		state.dieCount = dice.length - this.value.length;
		state.heldScore = total;
		delete state.dice;

		if (state.dieCount === 0) {
			state.dieCount = 6;
		}
	}
}

/**
 * Score takes a player index and a value which is a number representing the number of
 * points that a player has accrued by rolling and holding dice. Once a player has
 * scored, the active player is passed on to the next player and the players total score
 * is updated.
 *
 * If a player's score matches or exceeds 10,000, then a turn countdown begins, allowing
 * all the remaining players to take one more turn before the game ends.
 */
export class Score implements GameEvent {
	kind: GameEventKind.Score;
	player: number;
	value: number;

	constructor(event: GameEventData) {
		const player = throwIfNoPlayer(event);
		const value = throwIfNoValueNumber(event);

		this.kind = GameEventKind.Score;
		this.player = player;
		this.value = value;
	}

	run(state: State) {
		const { dieCount, heldScore, scores } = state;
		const playerCount = scores?.length;

		if (!playerCount) {
			throw new Error("there are no players");
		}

		throwIfGameOver(state);
		throwIfWrongTurn(state, this.player);

		if (dieCount === 6) {
			throw new Error("player must roll when they have six free dice");
		}

		if (this.value !== state.heldScore) {
			throw new Error("player score must match the held score");
		}

		// It's safe to tell the compiler that the player is not undefined because of the
		// check above.
		state.scores![this.player] += heldScore!;

		// Increment the index of the active player, circling back to 1 if the player
		// who just scored was the last player in the array.
		state.playing =
			playerCount - 1 === this.player
				? (state.playing = 0)
				: (state.playing = this.player + 1);

		state.dieCount = 6;
		delete state.heldScore;

		// If there is an active turn countdown, then now is the time to decrement it,
		// otherwise, if the player has crossed 10,000, then it's time to start the
		// turn countdown.
		if (state.turnCountdown !== undefined) {
			state.turnCountdown--;
		} else if (state.scores![this.player] >= GAME_SCORE_THRESHOLD) {
			state.turnCountdown = playerCount - 1;
		}
	}
}

function scorePips(count: number, pips: number) {
	if (count < 3) {
		// If not a three of a kind, return the raw dice value...
		return pipScore(pips) * count;
	}

	// ...otherwise, this is a three or more of a kind.
	if (pips === 1) {
		// Ones are treated as 10s.
		pips = 10;
	}

	// Three of a kind is the pips times 100, times two for each additional rolled dice.
	return pips * 100 * (count - 2);
}

/**
 * TODO: There are variations to the rules of ten thousand. It would be interesting if
 *       code like this could actually come from a setting somewhere.
 */
function pipScore(pips: number) {
	switch (pips) {
		case 1:
			return 100;
		case 5:
			return 50;
		default:
			return 0;
	}
}

function throwIfNoPlayer(event: GameEventData): number {
	if (event.player === undefined) {
		throw new Error("missing player index");
	}

	return event.player;
}

function throwIfNoValueNumber({ value }: GameEventData): number {
	if (typeof value !== "number") {
		throw new Error("missing value");
	}

	return value;
}

function throwIfNoValueArray({ value }: GameEventData): number[] {
	if (!Array.isArray(value)) {
		throw new Error("missing value array");
	}

	return value;
}

function throwIfGameOver(state: State) {
	if (state.turnCountdown === 0) throw new Error("the game is over");
}

function throwIfWrongTurn(state: State, player: number) {
	if (state.playing === undefined || state.playing !== player)
		throw new Error("it's not this player's turn");
}
