import type { RequestEvent } from "@sveltejs/kit";
import jwt from "jsonwebtoken";
import { type Method, type RouteAuthRule } from "./routeAuth";
import type { Listing } from "$lib/Listing";
import type { LoginData } from "$lib/Login";
import { isServerJwtPayload, type ServerJwtPayload } from "./ServerJwtPayload";

export type LocalCredentials = (
	| { kind: "Basic"; payload: { username: string; password: string } }
	| {
			kind: "Bearer";
			payload: ServerJwtPayload;
			role: string;
	  }
	| { kind: "None" }
) & { role: string };

export enum AuthorizationResult {
	Allowed,
	Denied,
	Unauthenticated,
}

export async function createToken(listing: Listing<LoginData>, secret: string) {
	const serverPayload: ServerJwtPayload = {
		sub: listing.id,
		username: listing.data.username,
		role: listing.data.role,
	};

	return await jwt.sign(serverPayload, secret, {
		expiresIn: "1d",
	});
}

export async function authenticate(
	event: RequestEvent,
	jwtSecret: string,
): Promise<LocalCredentials | null> {
	const authHeader = event.request.headers.get("authorization");
	let tokenRole: string;
	let tokenDesc: LocalCredentials;

	// get token kind and role from the header
	if (!authHeader) {
		// This is a stranger: they have no token and they will be assigned the default
		// role.
		tokenRole = "default";
		tokenDesc = { kind: "None", role: tokenRole };
	} else {
		const [kind, token] = authHeader.split(" ");

		if (kind === "Bearer") {
			// The role can be derived from the JWT token.
			const payload = await jwt.verify(token, jwtSecret);
			if (!isServerJwtPayload(payload)) {
				return null;
			}

			tokenRole = payload.role;
			tokenDesc = {
				kind: "Bearer",
				payload,
				role: tokenRole,
			};

			if (!tokenRole) {
				// Something has gone wrong: I should not have issued a token without a
				// role.
				return null;
			}
		} else if (kind === "Basic") {
			const decoded = Buffer.from(token, "base64").toString("ascii");
			const [username, password] = decoded.split(":");

			tokenRole = "default";

			tokenDesc = { kind: "Basic", payload: { username, password }, role: tokenRole };
		} else {
			// Something the server doesn't recognize was passed as the token kind.
			return null;
		}
	}

	return tokenDesc;
}

export function isAuthorized(
	roleRules: Record<string, RouteAuthRule[]>,
	method: Method,
	path: string,
	creds: LocalCredentials,
): AuthorizationResult {
	// Determine if the role has the ability to hit this endpoint with this method
	// and the given token.
	const parts = breakupPath(path);
	const { role: tokenRole, kind: tokenKind } = creds;
	const rules = roleRules[tokenRole];

	if (parts[0] !== "api") {
		// All the routes for the backed server are prefixed with api, and the backed
		// server is where all the authorization is going to happen so we will ignore
		// calls to the frontend.
		return AuthorizationResult.Allowed;
	}

	let hasMatchingAllow = false;
	for (const rule of rules) {
		if (matchesRequest(parts, method, rule)) {
			if (tokenKind !== (rule.tokenKind ?? "Bearer")) {
				return AuthorizationResult.Unauthenticated;
			} else if (rule.action === "deny") {
				// if a request matches any deny rule, then it is denied, regardless
				// of whether or not it also matches an allow rule.
				return AuthorizationResult.Denied;
			} else if (rule.action === "allow") {
				hasMatchingAllow = true;
			}
		}
	}

	if (hasMatchingAllow) return AuthorizationResult.Allowed;
	if (tokenKind === "Bearer") return AuthorizationResult.Denied;
	else return AuthorizationResult.Unauthenticated;
}

function breakupPath(path: string): string[] {
	// remove leading and trailing /
	if (path[0] === "/") path = path.slice(1);
	if (path[path.length - 1] === "/") path = path.slice(0, path.length - 1);

	return path.split("/");
}

function matchesRequest(
	requestParts: string[],
	method: Method,
	{ endpoint, methods }: RouteAuthRule,
): boolean {
	if (!methods.includes("*") && !methods.includes(method)) {
		return false;
	}

	const ruleParts = breakupPath(endpoint);
	for (let i = 0; i < ruleParts.length; i++) {
		const rulePart = ruleParts[i];
		const reqPart = requestParts[i];

		if (!reqPart) return false;

		if (rulePart[0] === "[" && rulePart[rulePart.length - 1] === "]") {
			// This part of the path represents an ID.
			continue;
		}

		if (i == ruleParts.length - 1 && rulePart === "*") {
			// Rule has a wildcard, anything after it automatically matches.
			return true;
		}

		if (rulePart !== reqPart) {
			return false;
		}
	}

	// Finished comparing without seeing a wildcard, meaning the requested path
	// and the path rules should also be the same length.
	return requestParts.length === ruleParts.length;
}
