import type { Id } from "$lib/Id";
import type { Listing } from "$lib/Listing";
import {
	MongoClient,
	ObjectId,
	ServerApiVersion,
	type Document,
	type WithId,
} from "mongodb";

type ListingFromMongo = Omit<Listing, "id"> & { _id: ObjectId };

export enum ServerCollections {
	Games = "games",
	Logins = "logins",
	Turns = "turns",
}

const uri = `mongodb://127.0.0.1:27017`;
const DATABASE = "ten-thousand";
let cachedClient: MongoClient | null = null;

async function getClient() {
	if (cachedClient !== null) return cachedClient;

	const c = new MongoClient(uri, {
		serverApi: {
			version: ServerApiVersion.v1,
			strict: true,
			deprecationErrors: true,
		},
	});

	await c.connect();
	return c;
}

export async function writeNewListing(col: ServerCollections, listing: Listing) {
	const client = await getClient();

	await client.db(DATABASE).collection(col).insertOne(fixListingForMongo(listing));
}

export async function writeUpdatedListing(col: ServerCollections, listing: Listing) {
	const client = await getClient();

	await client
		.db(DATABASE)
		.collection(col)
		.replaceOne({ _id: new ObjectId(listing.id) }, fixListingForMongo(listing));
}

export async function listCollection<T>(
	col: ServerCollections,
	dataGuard: (target: unknown) => target is Listing<T>,
) {
	const client = await getClient();
	const res = await client.db(DATABASE).collection(col).find();

	return (await res.toArray()).map((doc) => fixListingFromMongo(doc, dataGuard));
}

export async function readListingById<T>(
	col: ServerCollections,
	id: Id,
	dataGuard: (target: unknown) => target is Listing<T>,
) {
	const client = await getClient();
	const res = await client
		.db(DATABASE)
		.collection(col)
		.findOne({ _id: idFromString(id) });

	if (res === null) return null;
	return fixListingFromMongo(res, dataGuard);
}

export function idFromString(str: string) {
	return new ObjectId(str);
}

export function stringFromId(id: ObjectId) {
	return id.toString();
}

export async function readListingByQuery<T>(
	col: ServerCollections,
	query: object,
	dataGuard: (target: unknown) => target is Listing<T>,
): Promise<Listing<T> | null> {
	const client = await getClient();
	const res = await client.db(DATABASE).collection(col).findOne(query);

	if (res === null) return null;

	return fixListingFromMongo(res, dataGuard);
}

function fixListingForMongo(listing: Listing): ListingFromMongo {
	const { id, ...rest } = listing;
	return {
		_id: idFromString(id),
		...rest,
	};
}

function fixListingFromMongo<T>(
	target: WithId<Document>,
	dataGuard: (target: unknown) => target is Listing<T>,
): Listing<T> {
	const { _id, ...rest } = target;
	const adjusted = { id: stringFromId(_id), ...rest };

	if (!dataGuard(adjusted)) {
		throw new Error("the returned document does not conform to the provided type");
	}

	return adjusted;
}
