import { describe, expect, it } from "vitest";
import * as jwt from "jsonwebtoken";
import type { Listing } from "$lib/Listing";
import type { LoginData } from "$lib/Login";
import { createId } from "$lib/Id";
import {
	authenticate,
	AuthorizationResult,
	createToken,
	isAuthorized,
	type LocalCredentials,
} from "$lib/server/auth";
import type { Cookies, RequestEvent } from "@sveltejs/kit";
import type { Method, RouteAuthRule } from "$lib/server/routeAuth";

type TestData = {
	heading: string;
	conditions: {
		token: LocalCredentials;
		method: Method;
		path: string;
	};
	expectations: {
		result: AuthorizationResult;
	};
};

const JWT_SECRET = "server-secret";

const loginListing: Listing<LoginData> = {
	id: createId(),
	createdAt: new Date().toString(),
	modifiedAt: new Date().toString(),
	deleted: false,
	data: {
		username: "Somebody Important",
		password: "some-password-hash",
		role: "test-role",
	},
};

function getEvent(overrides: Partial<RequestEvent> = {}): RequestEvent {
	const event: RequestEvent = {
		cookies: {} as Cookies,
		fetch: async (): Promise<Response> => {
			return new Response();
		},
		getClientAddress: () => "",
		locals: { user: {} },
		params: {},
		platform: undefined,
		request: new Request(new URL("https://localhost/api")),
		route: { id: "" },
		setHeaders: () => {},
		url: new URL("https://localhost/api"),
		isDataRequest: false,
		isSubRequest: false,
	};

	return { ...event, ...overrides };
}

describe("auth", () => {
	describe("createToken", () => {
		it("should create a json web token", async () => {
			const token = await createToken(loginListing, "some-server-secret");

			expect(jwt.decode(token)).to.not.be.null;
		});
	});

	describe("authenticate", () => {
		it("should return a default/'None' object when no auth header is present", async () => {
			const event = getEvent();

			expect(await authenticate(event, JWT_SECRET)).to.deep.equal({
				kind: "None",
				role: "default",
			});
		});

		it("should return a 'Bearer' object with the correct user role when a valid jwt token is passed", async () => {
			const token = await createToken(loginListing, JWT_SECRET);
			const request = new Request("http://localhost/api", {
				headers: [["Authorization", `Bearer ${token}`]],
			});
			const event = getEvent({ request });
			const tokenDesc = await authenticate(event, JWT_SECRET);

			expect(tokenDesc).property("kind", "Bearer");
			const { payload, ...rest } = tokenDesc as {
				kind: "Bearer";
				role: string;
				payload: jwt.JwtPayload;
			};

			expect(payload).property("sub", loginListing.id.toString());
			expect(payload).property("username", loginListing.data.username);
			expect(payload).property("role", loginListing.data.role);
			expect(rest).to.deep.equal({
				kind: "Bearer",
				role: loginListing.data.role,
			});
		});

		it("should return null if a JWT token payload is a string", async () => {
			const token = await jwt.sign("I'm just a dumb string!", JWT_SECRET);

			const request = new Request("http://localhost/api", {
				headers: [["Authorization", `Bearer ${token}`]],
			});

			const event = getEvent({ request });
			const tokenDesc = await authenticate(event, JWT_SECRET);

			expect(tokenDesc).to.be.null;
		});

		it("should return null if a JWT token payload is missing a role", async () => {
			const token = await jwt.sign(
				{
					sub: loginListing.id,
					username: loginListing.data.username,
				},
				JWT_SECRET,
				{
					expiresIn: "1d",
				},
			);

			const request = new Request("http://localhost/api", {
				headers: [["Authorization", `Bearer ${token}`]],
			});

			const event = getEvent({ request });
			const tokenDesc = await authenticate(event, JWT_SECRET);

			expect(tokenDesc).to.be.null;
		});

		it("should return a default/'Basic' object when a Basic header is provided", async () => {
			const user = "someuser";
			const pass = "somepass";
			const token = Buffer.from(`${user}:${pass}`).toString("base64");
			const request = new Request("http://localhost/api", {
				headers: [["Authorization", `Basic ${token}`]],
			});

			const event = getEvent({ request });
			const tokenDesc = await authenticate(event, JWT_SECRET);

			expect(tokenDesc).to.deep.equal({
				kind: "Basic",
				payload: {
					username: user,
					password: pass,
				},
				role: "default",
			});
		});

		it("should return null if the token type is nonesense", async () => {
			const request = new Request("http://localhost/api", {
				headers: [["Authorization", "Nonesense tokenbody"]],
			});

			const event = getEvent({ request });
			const tokenDesc = await authenticate(event, JWT_SECRET);

			expect(tokenDesc).to.be.null;
		});
	});

	describe("isAuthorized", () => {
		const authRules: Record<string, RouteAuthRule[]> = {
			alternativeRole: [
				{
					action: "allow",
					methods: ["*"],
					endpoint: "/api/thing",
				},
			],
			superUser: [
				{
					action: "allow",
					methods: ["GET", "POST", "DELETE", "PUT"],
					endpoint: "*",
				},
			],
			tokenBearer: [
				// user is explicitely not allowed to use PATCH
				{ action: "deny", methods: ["PATCH"], endpoint: "*" },

				// user is allowed to POST to resources
				{
					action: "allow",
					methods: ["POST"],
					endpoint: "/api/resource",
					tokenKind: "Bearer",
				},

				// user is allowed to use any method on a single resource
				{ action: "allow", methods: ["*"], endpoint: "/api/resource/[resourceid]" },

				// user is allowed to GET or DELETE any resource nested under resource
				{
					action: "allow",
					methods: ["GET", "DELETE"],
					endpoint: "/api/resource/[resourceid]/*",
				},
			],
			tokenNone: [
				// user is allowed to hit an endpoint with a "None" token
				{ action: "allow", methods: ["POST"], endpoint: "/api/user", tokenKind: "None" },
			],
			tokenBasic: [
				// user is allowed to hit an endpoint with a "Basic token"
				{
					action: "allow",
					methods: ["POST"],
					endpoint: "/api/token",
					tokenKind: "Basic",
				},
			],
		};

		const tokenBearer = { kind: "Bearer", role: "tokenBearer", payload: {} } as const;

		const tests: TestData[] = [
			{
				heading: "should return allowed when a user calls an explicitly allowed path",
				conditions: {
					token: tokenBearer,
					method: "POST",
					path: "/api/resource",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading: "should match paths based on the role provided in the token",
				conditions: {
					token: { kind: "Bearer", role: "alternativeRole", payload: {} },
					method: "POST",
					path: "/api/thing",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading: "should ignore paths that only exist on other roles",
				conditions: {
					token: { kind: "Bearer", role: "alternativeRole", payload: {} },
					method: "POST",
					path: "/api/resource",
				},
				expectations: {
					result: AuthorizationResult.Denied,
				},
			},
			{
				heading: "should return unauthenticated when a user has the wrong token type",
				conditions: {
					token: {
						kind: "Basic",
						role: "tokenBearer",
						payload: { username: "Mr. Sneaky", password: "something-secret" },
					},
					method: "POST",
					path: "/api/resource",
				},
				expectations: {
					result: AuthorizationResult.Unauthenticated,
				},
			},
			{
				heading: "should return allowed when a rule opens a path with a wildcard method",
				conditions: {
					token: tokenBearer,
					method: "PUT",
					path: "/api/resource/some-resource-id",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading: "should return denied when the endpoint is valid, but the method is not",
				conditions: {
					token: tokenBearer,
					method: "GET",
					path: "/api/resource",
				},
				expectations: {
					result: AuthorizationResult.Denied,
				},
			},
			{
				heading: "should return allowed when matching a wildcard endpoint",
				conditions: {
					token: {
						kind: "Bearer",
						role: "superUser",
						payload: {},
					},
					method: "PUT",
					path: "/api/magicresource",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading:
					"should return denied wehen matching a wildcard endpoint with an invalid method",
				conditions: {
					token: {
						kind: "Bearer",
						role: "superUser",
						payload: {},
					},
					method: "PATCH",
					path: "/api/magicresource",
				},
				expectations: {
					result: AuthorizationResult.Denied,
				},
			},
			{
				heading: "should return allowed when matching a nested endpoint wildcard",
				conditions: {
					token: tokenBearer,
					method: "GET",
					path: "/api/resource/some-resource-id/nested-resource",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading:
					"should return return allowed when matching a resource deeply " +
					"nested under an endpoint wildcard",
				conditions: {
					token: tokenBearer,
					method: "GET",
					path:
						"/api/resource/some-resource-id/nested-resource/" +
						"more-nested-resource/more-nested-id",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading:
					"should not match a path rule if the request path ends before the " +
					"rule path's endpoint wildcard",
				conditions: {
					token: tokenBearer,
					method: "GET",
					path: "/api/resource",
				},
				expectations: {
					result: AuthorizationResult.Denied,
				},
			},
			{
				heading:
					"should return denied when a 'denied' rule and an 'allowed' rule both match",
				conditions: {
					token: tokenBearer,
					method: "PATCH",
					path: "/api/resource/some-resource-id",
				},
				expectations: {
					result: AuthorizationResult.Denied,
				},
			},
			{
				heading: "should return denied when a user calls an unmentioned path",
				conditions: {
					token: tokenBearer,
					method: "GET",
					path: "/api/unmentioned",
				},
				expectations: {
					result: AuthorizationResult.Denied,
				},
			},
			{
				heading:
					"should return unauthenticated when a user calls an unmentioned " +
					"path with a 'None' token",
				conditions: {
					token: { kind: "None", role: "tokenBearer" },
					method: "GET",
					path: "/api/unmentioned",
				},
				expectations: {
					result: AuthorizationResult.Unauthenticated,
				},
			},
			{
				heading:
					"should return unauthenticated when a user calls an unmentioned " +
					"path with a 'Basic' token",
				conditions: {
					token: {
						kind: "Basic",
						role: "tokenBearer",
						payload: { username: "someuser", password: "super-secure" },
					},
					method: "GET",
					path: "/api/unmentioned",
				},
				expectations: {
					result: AuthorizationResult.Unauthenticated,
				},
			},
			{
				heading: "allows a user to hit any endpoint that doesn't start with api",
				conditions: {
					token: {
						kind: "None",
						role: "default",
					},
					method: "GET",
					path: "/resource",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading: "correctly matches a route with a trailing /",
				conditions: {
					token: tokenBearer,
					method: "POST",
					path: "/api/resource/",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
			{
				heading: "corretly matches a route without a leading /",
				conditions: {
					token: tokenBearer,
					method: "POST",
					path: "api/resource",
				},
				expectations: {
					result: AuthorizationResult.Allowed,
				},
			},
		];

		for (const { heading, conditions, expectations } of tests) {
			it(heading, () => {
				const { token, method, path } = conditions;
				const { result } = expectations;

				expect(isAuthorized(authRules, method, path, token)).to.equal(result);
			});
		}
	});
});
