import { createId } from "$lib/Id";
import { isListing, type Listing } from "$lib/Listing";
import { describe, expect, it } from "vitest";

let spiedValue: unknown;

const passingDataGuard = (target: unknown): target is any => true;
const failingDataGuard = (target: unknown): target is any => false;
const spyingDataGuard = (target: unknown): target is any => {
	spiedValue = target;
	return true;
};

describe("Listing", () => {
	describe("isListing", () => {
		it("should return false if a required property is missing", () => {
			const fullListing: Listing = {
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString(),
				deleted: false,
				data: {},
			};

			let listing: Partial<Listing> = { ...fullListing };
			delete listing.id;
			expect(isListing(listing)).to.be.false;

			listing = { ...fullListing };
			delete listing.createdAt;
			expect(isListing(listing)).to.be.false;

			listing = { ...fullListing };
			delete listing.deleted;
			expect(isListing(listing)).to.be.false;

			listing = { ...fullListing };
			delete listing.modifiedAt;
			expect(isListing(listing)).to.be.false;

			listing = { ...fullListing };
			delete listing.data;
			expect(isListing(listing)).to.be.false;
		});

		it("should return false if there is an extra property", () => {
			const fullListing = {
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString(),
				deleted: false,
				extra: "property",
				data: {},
			};

			expect(isListing(fullListing)).to.be.false;
		});

		it("should return true if all the required properties are present", () => {
			const fullListing = {
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString() as string | null,
				deleted: false,
				data: {},
			};

			expect(isListing(fullListing)).to.be.true;

			fullListing.modifiedAt = null;
			expect(isListing(fullListing)).to.be.true;
		});

		it("should return true if properties are present and the data guard returns true", () => {
			const fullListing = {
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString() as string | null,
				deleted: false,
				data: {},
			};

			expect(isListing(fullListing, passingDataGuard)).to.be.true;
		});

		it("should return false if properties are present and the data guard returns false", () => {
			const fullListing = {
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString() as string | null,
				deleted: false,
				data: {},
			};

			expect(isListing(fullListing, failingDataGuard)).to.be.false;
		});

		it("should pass the data object to the provided data guard", () => {
			const fullListing = {
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString() as string | null,
				deleted: false,
				data: { am: "the test object" },
			};

			expect(isListing(fullListing, spyingDataGuard)).to.be.true;
			expect(spiedValue).to.deep.equal(fullListing.data);
		});
	});
});
