import { describe, it } from "vitest";
import { equal, ok } from "node:assert/strict";
import { hasProperty, hasOnlyKeys } from "$lib/validation";

describe("validation", () => {
	describe("hasProperty", () => {
		it("should return false if the property is undefined", () => {
			const target = { some: "property" };
			const result = hasProperty(target, "important", "string");

			equal(result, false);
		});

		it("should return false if passed a non-object", () => {
			const target = 45;
			const result = hasProperty(target, "important", "string");

			equal(result, false);
		});

		it("should return false if passed null", () => {
			const target = null;
			const result = hasProperty(target, "important", "string");

			equal(result, false);
		});

		it("should return false if passed undefined", () => {
			const target = undefined;
			const result = hasProperty(target, "important", "string");

			equal(result, false);
		});

		it("should return false if the property is of the wrong type", () => {
			const target = { important: 45 };
			const result = hasProperty(target, "important", "string");

			equal(result, false);
		});

		it("should return true if the property is the correct type", () => {
			const target = {
				first: "string",
				second: 2,
				third: false,
				fourth: null,
				fifth: { something: "important" },
				sixth: ["one", "two"],
			};

			ok(hasProperty(target, "first", "string"));
			ok(hasProperty(target, "second", "number"));
			ok(hasProperty(target, "third", "boolean"));
			ok(hasProperty(target, "fourth", "null"));
			ok(hasProperty(target, "fifth", "object"));
			ok(hasProperty(target, "sixth", "array"));
		});

		it("should return false if passed an array type and the property isn't an array", () => {
			const target = {
				arr: "not array",
			};

			equal(hasProperty(target, "arr", "string[]"), false);
		});

		it("should return false if the defined array contains a non-matching element", () => {
			const target = {
				arr: ["I", "was", "born", "in", 1989],
			};

			equal(hasProperty(target, "arr", "string[]"), false);
		});

		it("should return true if all the elements in a defined array match", () => {
			const target = {
				arr: ["I", "was", "born", "in", "1989"],
			};

			ok(hasProperty(target, "arr", "string[]"));
		});

		it("should return true if all the elements in a defined array match one of multiple types", () => {
			const target = {
				arr: ["I", "was", "born", "in", 1989],
			};

			ok(hasProperty(target, "arr", "(string|number)[]"));
		});

		it("should return true if type is null but property is nullable", () => {
			const target = {
				nullable: null,
			};

			ok(hasProperty(target, "nullable", "string", true));
		});
	});

	describe("hasOnlyKeys", () => {
		it("returns false if the target is not an object", () => {
			equal(hasOnlyKeys(45, []), false);
		});

		it("returns false has extra properties", () => {
			const target = {
				one: "one",
				two: "two",
				three: "three",
			};

			const keys = ["one", "two"];

			equal(hasOnlyKeys(target, keys), false);
		});

		it("should return true if the target has only the provided keys", () => {
			const target = {
				one: "one",
				two: "two",
				three: "three",
			};

			const keys = ["one", "two", "three"];

			ok(hasOnlyKeys(target, keys));
		});

		it("should return true if the target has only a subset of the provided keys", () => {
			const target = {
				one: "one",
			};

			const keys = ["one", "two", "three"];

			ok(hasOnlyKeys(target, keys));
		});
	});
});
