// hasProperty takes a target, which is unknown, propertyName and propertyType, which
// are string and isNullable, which is a boolean. If the property exists on the target,
// and the type of the property value matches the string passes (as determined using the
// typeof keyword) then hasProperty returns true. An important exception: if the user
// passes null or an array and a propertyType of "object", hasProperty will return false.
// To check if something is an array, pass "array" and to check if something is null, pass
// "null" as the property type. If hasProperty receives a null value, it will still return
// true as long as "isNullable" has been set to true (it defaults to false).
//
// hasProperty can also receive a specific array type which resembles defining arrays of
// items in TypeScript: e.g. string[] is an array of string, (string|number)[] is an array of
// strings or numbers. It does not recognize the Array<string> syntax.
export function hasProperty(
	target: unknown,
	propertyName: string,
	propertyType: string | ((target: unknown) => boolean),
	isNullable: boolean = false
): boolean {
	if (target === null || target === undefined) return false;

	const p = (target as any)[propertyName];

	if (p === undefined) {
		return false;
	}

	if (typeof propertyType === "function") {
		return propertyType(p);
	}

	if (p === null) {
		return propertyType === "null" || isNullable;
	}

	if (propertyType === "array") {
		return Array.isArray(p);
	}

	// TODO: this logic won't work on arrays of array, for instance, string[][]
	if (propertyType.substring(propertyType.length - 2) === "[]") {
		const types = parseArrayType(propertyType);

		if (!Array.isArray(p)) {
			return false;
		}

		for (const item of p) {
			let match = false;

			for (const type of types) {
				if (typeof item === type) {
					match = true;
					break;
				}
			}

			if (!match) {
				return false;
			}
		}

		return true;
	}

	return typeof p === propertyType;
}

// hasOnlyKeys takes a target, which is unknown, and keys, which is a string array.
// if the target has ONLY own properties that are listed on the string array, then
// hasOnlyKeys returns true. If there is any extra properties, it returns false.
//
// hasOnlyKeys does not make sure that the keys provided exist on the target, only
// that keys not provided do not exist.
export function hasOnlyKeys(target: unknown, keys: string[]) {
	if (typeof target !== "object") return false;

	const keySet = new Set(keys);
	for (const key in target) {
		if (target.hasOwnProperty(key)) {
			if (!keySet.has(key)) return false;
		}
	}

	return true;
}

function parseArrayType(propertyType: string): string[] {
	if (propertyType[0] === "(") {
		const typesString = propertyType.substring(1, propertyType.length - 3);
		return typesString.split("|");
	} else {
		return [propertyType.substring(0, propertyType.length - 2)];
	}
}
