import type { Cookies, RequestEvent } from "@sveltejs/kit";
import { describe, it, expect, afterEach } from "vitest";
import * as auth from "../lib/server/auth";
import { getHandleFn } from "../hooks.server";
import { createId } from "$lib/Id";

let events: RequestEvent[] = [];

// Mock RequestEvent data that can be passed into the handle function. It doesn't matter
// that most of these fields are incoherent, the only things that will be tested here are
// that this event is passed to the isAuthorized function, and that the locas are
// populated.
const event: RequestEvent = {
	cookies: {} as Cookies,
	fetch: async (): Promise<Response> => {
		return new Response();
	},
	getClientAddress: () => "",
	locals: { user: {} },
	params: {},
	platform: undefined,
	request: new Request(new URL("https://localhost/api")),
	route: { id: "" },
	setHeaders: () => {},
	url: new URL("https://localhost/api"),
	isDataRequest: false,
	isSubRequest: false,
};

const resolve = async (event: RequestEvent): Promise<Response> => {
	events.push(event);
	return new Response();
};

describe("getHandleFn", () => {
	afterEach(() => {
		event.locals.user = {};
		event.request.headers.delete("authorization");
		events = [];
	});

	it("returns unauthorized response if caller isn't properly authenticated", async () => {
		event.request.headers.set("authorization", "Nonesense Token");
		const res = await getHandleFn("server-secret")({ event, resolve });
		expect(res.status).to.equal(401);
	});

	it("returns unauthorized response if caller is missing required auth header", async () => {
		const res = await getHandleFn("server-secret")({ event, resolve });
		expect(res.status).to.equal(401);
	});

	it("returns forbidden response if caller isn't authorized", async () => {
		// This is a weird scenario, but it does reflect the way I expect this to work.
		// Svelte Kit does not seem to provide me with a tool that I can use to authorize
		// users in one place. I would have to check their role at the start of each
		// endpoint function (yuck). The endpoint below doesn't exist, but it will still
		// return 403 instead of 404 because the auth check happens before the route
		// matching, and the user isn't authorized to hit this nonesense endpoint.
		const ev: RequestEvent = {
			...event,
			url: new URL("https://localhost/api/some/secret/route"),
			request: new Request("https://localhost/api/some/secret/route"),
		};

		const token = await auth.createToken(
			{
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString(),
				deleted: false,
				data: {
					password: "somethin' secret!",
					username: "Mr. Man",
					role: "default",
				},
			},
			"server-secret",
		);

		ev.request.headers.set("authorization", `Bearer ${token}`);

		const res = await getHandleFn("server-secret")({ event: ev, resolve });
		expect(res.status).to.equal(403);
	});

	it("attaches information about the credentials to the event and resolves when the user is authorized", async () => {
		const ev: RequestEvent = {
			...event,
			url: new URL("https://localhost/api/some/secret/route"),
			request: new Request("https://localhost/api/some/secret/route"),
		};

		const token = await auth.createToken(
			{
				id: createId(),
				createdAt: new Date().toString(),
				modifiedAt: new Date().toString(),
				deleted: false,
				data: {
					password: "somethin' secret!",
					username: "Mr. Man",
					role: "player",
				},
			},
			"server-secret",
		);

		ev.request.headers.set("authorization", `Bearer ${token}`);

		const res = await getHandleFn("server-secret")({ event: ev, resolve });
		expect(res.status).to.equal(200);

		const user: auth.LocalCredentials = events[0].locals.user;

		expect(user.role).to.equal("player");
		expect(user.kind).to.equal("Bearer");
	});
});
